(*
  TPLib: Tropical Polyhedra Library

  Copyright (C) 2009-2013 Xavier ALLAMIGEON (xavier.allamigeon at inria.fr)

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*)

let randobj = Gmp_random.init_default ();

module Mlgmp_int =
struct
  type t = Mpz.t

  let zero = Mpz.of_int 0
  let max x y = if Mpz.cmp x y >= 0 then x else y
  let min x y = if Mpz.cmp x y >= 0 then y else x
  let add x y =
    let res = Mpz.init () in
    Mpz.add res x y;
    res
  let neg x =
    let res = Mpz.init () in
    Mpz.neg res x;
    res
  let mul x y =
    let res = Mpz.init () in
    Mpz.mul res x y;
    res
  let div x y =
    let res = Mpz.init () in
    Mpz.cdiv_q res x y;
    res
  let pow x y =
    let res = Mpz.init () in
    Mpz.pow_ui res x y;
    res
  let compare = Mpz.cmp
  let random n =
    let res = Mpz.init () in
    Gmp_random.Mpz.urandomb res randobj n;
    res
  let of_int = Mpz.of_int
  let of_string = Mpz.of_string
  let to_string = Mpz.get_str ~base:10
end

module Mlgmp_rat =
struct
  type t = Mpq.t

  let zero = Mpq.of_int 0
  let one = Mpq.of_int 1
  let max x y = if Mpq.cmp x y >= 0 then x else y
  let min x y = if Mpq.cmp x y >= 0 then y else x
  let add x y =
    let res = Mpq.init () in
    Mpq.add res x y;
    res
  let neg x =
    let res = Mpq.init () in
    Mpq.neg res x;
    res
  let mul x y =
    let res = Mpq.init () in
    Mpq.mul res x y;
    res
  let div x y =
    let res = Mpq.init () in
    Mpq.div res x y;
    res

  let pow x n =
    let rec pow_aux x n =
      assert (n >= 0);
      if n = 0 then one
      else
        let y = pow_aux x (n/2) in
        let z = Mpq.init () in
        Mpq.mul z y y;
        if n mod 2 != 0 then
          Mpq.mul z x z;
        z
    in
    if n >= 0 then
      pow_aux x n
    else
      pow_aux (Mpq.inv x x; x) (-n)

  let compare = Mpq.cmp
  let random n =
    let num = Mpz.init () in
    let den = Mpz.init () in
    Gmp_random.Mpz.urandomb num randobj n;
    Gmp_random.Mpz.urandomb den randobj n;
    let res = Mpq.init () in
    Mpq.div res (Mpq.init_set_z num) (Mpq.add res one (Mpq.init_set_z den); res);
    res
  let of_int = Mpq.of_int
  let of_string = Mpq.of_string
  let to_string = Mpq.to_string
end

let _ =
  Numeric_plugin.add_module "mlgmp_int" (module Mlgmp_int: Numeric_plugin.T);
  Numeric_plugin.add_module "mlgmp_rat" (module Mlgmp_rat: Numeric_plugin.T)
